# -*- coding: utf-8 -*-
__version__ = "0.0.1"

import logging
import importlib
import re
import json
from abc import ABC, abstractmethod
from environs import Env

try:
    import yaml
except ImportError:
    pass  # will fail if using yaml config_file


logger = logging.getLogger()

env = Env()

default_app = None


def eval_globals():
    """Global modules that might be used in expressions"""
    import math
    import datetime
    return {"datetime": datetime, "math": math, "json": json}


class EventApp:

    def __init__(self, name_prefix="", listeners=None):
        self.name_prefix = name_prefix or ""
        listeners = listeners or []
        self.listeners = []
        [self.add_listener(lis) for lis in listeners]

    def add_listener(self, listener):
        if not isinstance(listener, Listener):
            listener = Listener.from_dict(listener)
        self.listeners.append(listener)

    def trigger(self, event_name, event_data):
        event_name = self.name_prefix + event_name
        count = 0
        for lis in self.listeners:
            sent = lis.notify(event_name, event_data)
            if sent:
                count += 1
        return count


class Listener(ABC):
    _registry = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "kind"):
            Listener._registry[cls.kind] = cls

    @classmethod
    def get(cls, kind):
        if kind not in cls._registry:
            # Try import listener module
            try:
                importlib.import_module(f"eventisc.{kind}_listener")
            except ImportError:
                pass
        return cls._registry[kind]

    @classmethod
    def from_dict(cls, listener_def):
        listener_def = dict(listener_def)
        kind = listener_def.pop("kind")
        listener_cls = cls.get(kind)
        if "event_name_regex" in listener_def:
            listener_def["event_name"] = re.compile(listener_def.pop("event_name_regex"))
        if "filter" in listener_def:
            listener_def["filter"] = Filter.from_dict(listener_def["filter"])
        return listener_cls(**listener_def)

    def __init__(self, event_name, filter=None):
        self.event_name = event_name
        self.filter = filter

    def notify(self, event_name, event_data):
        if hasattr(self.event_name, "match"):  # is REGEX
            m = self.event_name.match(event_name)
            if not m:
                return False
        elif self.event_name != event_name:
            return False

        if self.filter and not self.filter(event_name, event_data):
            return False

        if env.bool("EVENTISC_DRYRUN"):
            return self._do_notify_dryrun(event_name, event_data)
        else:
            return self._do_notify(event_name, event_data)

    @abstractmethod
    def _do_notify(self, event_name, event_data):
        raise NotImplementedError

    def _do_notify_dryrun(self, event_name, event_data):
        logger.info("EVENTISC_DRYRUN=Y _do_notify(%s, %s)", event_name, event_data)
        return True

    @classmethod
    def format(self, value, event_name, event_data, **kwargs):
        """Formats a field that can be constant, proxy or format string of the event values"""
        if not isinstance(value, str) or '{' not in value:
            return value
        context_dict = dict(event_name=event_name, event_data=event_data, **kwargs)
        if value.startswith("{") and value.endswith("}") and '{' not in value[1:-1]:
            # It's a simple value, {variable}, use eval to keep type
            try:
                expr = compile(value[1:-1], __file__, "eval")
            except SyntaxError:
                return value.format(**context_dict)
            return eval(expr, eval_globals(), context_dict)
        return value.format(**context_dict)


class Filter(ABC):
    _registry = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "kind"):
            Filter._registry[cls.kind] = cls

    @classmethod
    def get(cls, kind):
        return cls._registry[kind]

    @classmethod
    def from_dict(cls, filter_def):
        filter_def = dict(filter_def)
        kind = filter_def.pop("kind")
        filter_cls = cls.get(kind)
        return filter_cls(**filter_def)

    @abstractmethod
    def __call__(self, event_name, event_data):
        raise NotImplementedError


class ExprFilter(Filter):
    kind = "expr"

    def __init__(self, expr):
        self.expr = compile(expr, __file__, "eval")

    def __call__(self, event_name, event_data):
        return bool(eval(self.expr, eval_globals(), {"event_name": event_name, "event_data": event_data}))


def read_config_file(config_file):
    if hasattr(config_file, "read"):
        # Assume json format
        format = "json"
    elif config_file.endswith(".yaml") or config_file.endswith(".yml"):
        format = "yaml"
        config_file = open(config_file, "rt")
    else:
        format = "json"
        config_file = open(config_file, "rt")

    if format == "json":
        config = json.load(config_file)
    elif format == "yaml":
        config = yaml.safe_load(config_file)

    return config.get("name_prefix", None), config.get("listeners", [])


def init_default_app(name_prefix=None, listeners=None, config_file=None):
    global default_app
    if name_prefix is None and listeners is None and config_file is None:
        # Try reading from env
        config_file = env.str("EVENTISC_CONFIG", None)
        if config_file is None:
            name_prefix = env.str("EVENTISC_NAME_PREFIX", "")
            listeners = []
        else:
            name_prefix, listeners = read_config_file(config_file)
    elif name_prefix is None and config_file is None:
        name_prefix = env.str("EVENTISC_NAME_PREFIX", "")

    default_app = create_app(name_prefix, listeners)
    return default_app


def create_app(name_prefix=None, listeners=None, config_file=None):
    if config_file is not None:
        name_prefix, listeners = read_config_file(config_file)

    return EventApp(name_prefix, listeners)


def get_current_app():
    global default_app

    if default_app is None:
        init_default_app()
    return default_app


def trigger(event_name, event_data):
    return get_current_app().trigger(event_name, event_data)



# -*- coding: utf-8 -*-
from unittest import TestCase
import eventisc
import eventisc.celery_listener

from unittest import mock


class TestCeleryListener(TestCase):
    def tearDown(self):
        eventisc.default_app = None

    @mock.patch.object(eventisc.celery_listener, "current_app")
    def test_celery_listener(self, celery_mock):
        app = eventisc.init_default_app(listeners=[
            {
                "event_name_regex": ".*",
                "kind": "celery", "queue": "cola", "task_name": "{event_name}",
                "task_kwargs": {"foo": "{event_data['foo']}", "fofofo": "Esto esta {event_data[foo]}",
                                "bar": 3, "barcito": "{event_data['bar']}"},
                "send_task_kargs": {"priority": 3, "serializer": "json"},
            }
        ])

        app.trigger('something', {"foo": "FFOO", "bar": 43})
        celery_mock.send_task.assert_called_once_with(
            'something', args=[],
            kwargs={'foo': 'FFOO', 'fofofo': 'Esto esta FFOO', 'bar': 3, 'barcito': 43},
            priority=3, queue='cola', serializer='json'
        )

    @mock.patch.object(eventisc.celery_listener, "current_app")
    def test_celery_default_kwargs(self, celery_mock):
        app = eventisc.init_default_app(listeners=[
            {
                "event_name_regex": ".*",
                "kind": "celery", "queue": "cola", "task_name": "event_receiver",
            }
        ])

        app.trigger('something', {"foo": "FFOO", "bar": 43})
        celery_mock.send_task.assert_called_once_with(
            'event_receiver', args=[],
            kwargs={'event_name': 'something', "event_data": {"foo": "FFOO", "bar": 43}},
            queue="cola",
        )

    @mock.patch.object(eventisc.celery_listener, "current_app")
    def test_celery_both(self, celery_mock):
        app = eventisc.init_default_app(listeners=[
            {
                "event_name_regex": ".*",
                "task_args": ["{event_name}"], "task_kwargs": "**event_data",
                "kind": "celery", "queue": "cola", "task_name": "event_receiver",
            }
        ])

        app.trigger('something', {"foo": "FFOO", "bar": 43})
        celery_mock.send_task.assert_called_once_with(
            'event_receiver', args=["something"],
            kwargs={"foo": "FFOO", "bar": 43},
            queue="cola",
        )



from urllib.parse import urlencode
import requests
from . import Listener


class HttpListener(Listener):
    kind = "http"

    def __init__(self, event_name, url, method="post", query_kwargs=None, data=None,
                 requests_kwargs=None, request_format="json", filter=None):
        super().__init__(event_name, filter)
        self.url = url
        self.method = method
        self.requests_kwargs = requests_kwargs or {}
        if "auth" in self.requests_kwargs:
            # Fix: auth parameter must be tuple and if read from yaml is list
            self.requests_kwargs["auth"] = tuple(self.requests_kwargs["auth"])
        self.query_kwargs = query_kwargs
        self.data = data
        self.request_format = request_format

    def _do_notify(self, event_name, event_data):
        if self.query_kwargs:
            query_kwargs = dict((k, self.format(v, event_name, event_data))
                                for (k, v) in self.query_kwargs.items())
            query_string = urlencode(query_kwargs)
        else:
            query_string = ""

        url = self.format(self.url, event_name, event_data, query_string=query_string)

        if self.data:
            data = dict((k, self.format(v, event_name, event_data))
                        for (k, v) in self.data.items())
        else:
            data = None

        request_method = getattr(requests, self.method)

        if self.request_format == "json":
            resp = request_method(url, json=data, **self.requests_kwargs)
        else:
            resp = request_method(url, data=data, **self.requests_kwargs)

        resp.raise_for_status()

        return True



# -*- coding: utf-8 -*-
from unittest import TestCase
import json
import responses
import eventisc


class TestHTTPListenerApp(TestCase):
    def tearDown(self):
        eventisc.default_app = None

    @responses.activate
    def test_http_listener(self):
        app = eventisc.init_default_app(listeners=[
            {
                "event_name_regex": ".*",
                "kind": "http",
                "url": "http://example.com/notify/{event_name}/?{query_string}",
                "query_kwargs": {"fifi": "{event_data['foo']}"},
                "data": {"foo": "{event_data['foo']}", "fofofo": "Esto esta {event_data[foo]}",
                         "bar": 3, "barcito": "{event_data['bar']}"},
                "requests_kwargs": {"headers": {"X-Foo": "sss"}},
            }
        ])

        responses.add(
            responses.POST,
            "http://example.com/notify/something/?fifi=FFOO",
        )

        app.trigger('something', {"foo": "FFOO", "bar": 43})

        request = responses.calls[0].request
        assert json.loads(request.body) == {
            "foo": "FFOO", "fofofo": "Esto esta FFOO",
            "bar": 3, "barcito": 43
        }
        assert request.headers["X-Foo"] == "sss"

    @responses.activate
    def test_auth_param(self):
        app = eventisc.init_default_app(listeners=[
            {
                "event_name_regex": ".*",
                "kind": "http",
                "url": "http://example.com/notify/{event_name}/",
                "data": {"foo": "{event_data['foo']}"},
                "requests_kwargs": {"auth": ["myuser", "mypassword"]},
            }
        ])

        responses.add(
            responses.POST,
            "http://example.com/notify/something_else/",
        )

        app.trigger('something_else', {"foo": "bar"})

        request = responses.calls[0].request
        assert json.loads(request.body) == {
            "foo": "bar"
        }
        assert request.headers["Authorization"] == "Basic bXl1c2VyOm15cGFzc3dvcmQ="



from . import Listener
from celery import current_app


class CeleryListener(Listener):
    kind = "celery"

    def __init__(self, event_name, queue, task_name, task_args=None, task_kwargs=None,
                 send_task_kargs={}, filter=None):
        super().__init__(event_name, filter)
        self.queue = queue
        self.task_name = task_name
        if task_args is None and task_kwargs is None:
            self.task_args = []
            self.task_kwargs = {"event_name": "{event_name}", "event_data": "{event_data}"}
        elif task_args is None:
            self.task_args = []
            self.task_kwargs = task_kwargs
        elif task_kwargs is None:
            self.task_args = task_args
            self.task_kwargs = {}
        else:
            self.task_args = task_args
            self.task_kwargs = task_kwargs
        self.send_task_kargs = send_task_kargs

    def _do_notify(self, event_name, event_data):
        queue = self.format(self.queue, event_name, event_data)
        task_name = self.format(self.task_name, event_name, event_data)
        args = [self.format(arg, event_name, event_data) for arg in self.task_args]
        if self.task_kwargs == "**event_data":
            kwargs = event_data
        else:
            kwargs = dict((k, self.format(v, event_name, event_data))
                          for (k, v) in self.task_kwargs.items())
        current_app.send_task(
            task_name,
            args=args, kwargs=kwargs,
            queue=queue,
            **self.send_task_kargs
        )
